"""
JSON-RPC version 2.0 protocol response implementation.
"""

import abc
import dataclasses as dc
import itertools as it
import operator as op
from typing import Any, ClassVar, Iterator, Optional

from pjrpc.common.typedefs import JsonRpcRequestIdT

from .common import UNSET, JsonT, MaybeSet
from .exceptions import DeserializationError, IdentityError, JsonRpcError


@dc.dataclass
class AbstractResponse(abc.ABC):
    """
    JSON-RPC 2.0 abstract response.
    """

    @classmethod
    @abc.abstractmethod
    def from_json(cls, data: JsonT, error_cls: type[JsonRpcError] = JsonRpcError) -> 'AbstractResponse':
        pass

    @abc.abstractmethod
    def to_json(self) -> JsonT:
        pass

    @property
    @abc.abstractmethod
    def is_success(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def is_error(self) -> bool:
        pass

    @abc.abstractmethod
    def unwrap_result(self) -> JsonT:
        pass

    @abc.abstractmethod
    def unwrap_error(self) -> JsonRpcError:
        pass


@dc.dataclass
class Response(AbstractResponse):
    """
    JSON-RPC 2.0 response.

    :param id: response identifier
    :param result: response result
    :param error: response error
    """

    version: ClassVar[str] = '2.0'

    id: Optional[JsonRpcRequestIdT] = None
    result: MaybeSet[JsonT] = UNSET
    error: MaybeSet[JsonRpcError] = UNSET

    @classmethod
    def from_json(cls, data: JsonT, error_cls: type[JsonRpcError] = JsonRpcError) -> 'Response':
        """
        Deserializes a response from json data.

        :param data: data the response to be deserialized from
        :param error_cls: error class
        :returns: response object
        :raises: :py:class:`pjrpc.common.exceptions.DeserializationError` if format is incorrect
        """

        try:
            if not isinstance(data, dict):
                raise DeserializationError("data must be of type dict")

            jsonrpc = data['jsonrpc']
            if jsonrpc != cls.version:
                raise DeserializationError(f"jsonrpc version '{data['jsonrpc']}' is not supported")

            request_id = data.get('id')
            if request_id is not None and not isinstance(request_id, (int, str)):
                raise DeserializationError("field 'id' must be of type integer or string")

            error_data = data.get('error', UNSET)
            error: MaybeSet[JsonRpcError]
            if error_data is not UNSET:
                if not isinstance(error_data, dict):
                    raise DeserializationError("error must be of type dict")
                error = error_cls.from_json(error_data)
            else:
                error = UNSET

            result = data.get('result', UNSET)
            if result is UNSET and error is UNSET:
                raise DeserializationError("'result' or 'error' fields must be provided")
            if result and error:
                raise DeserializationError("'result' and 'error' fields are mutually exclusive")

            return cls(id=request_id, result=result, error=error)
        except KeyError as e:
            raise DeserializationError(f"required field {e} not found") from e

    def __post_init__(self) -> None:
        assert self.result is not UNSET or self.error is not UNSET, "result or error argument must be provided"
        assert self.result is UNSET or self.error is UNSET, "result and error arguments are mutually exclusive"

    @property
    def is_success(self) -> bool:
        """
        Returns ``True`` if the response has succeeded.
        """

        return self.result is not UNSET

    @property
    def is_error(self) -> bool:
        """
        Returns ``True`` if the response has not succeeded.
        """

        return self.error is not UNSET

    def unwrap_result(self) -> JsonT:
        """
        Returns result. If result is not set raises and exception.
        """

        if self.is_error:
            raise self.unwrap_error()

        assert self.result is not UNSET, "result is not set"
        return self.result

    def unwrap_error(self) -> JsonRpcError:
        """
        Returns error. If error is not set raises and exception.
        """

        assert self.error is not UNSET, "error is not set"
        return self.error

    def to_json(self) -> JsonT:
        """
        Serializes the response to json data.

        :returns: json data
        """

        data: dict[str, JsonT] = {
            'jsonrpc': self.version,
            'id': self.id,
        }
        if self.result is not UNSET:
            data.update(result=self.result)
        if self.error is not UNSET:
            data.update(error=self.error.to_json())

        return data


@dc.dataclass
class BatchResponse(AbstractResponse):
    """
    JSON-RPC 2.0 batch response.

    :param responses: responses to be added to the batch
    """

    version: ClassVar[str] = '2.0'

    responses: MaybeSet[tuple[Response, ...]] = UNSET
    error: MaybeSet[JsonRpcError] = UNSET

    @classmethod
    def from_json(
            cls,
            data: JsonT,
            error_cls: type[JsonRpcError] = JsonRpcError,
            check_ids: bool = True,
    ) -> 'BatchResponse':
        """
        Deserializes a batch response from json data.

        :param data: data the response to be deserialized from
        :param error_cls: error class
        :param check_ids: check response ids for uniqueness
        :returns: batch response object
        """

        try:
            if isinstance(data, dict):
                jsonrpc = data['jsonrpc']
                if not isinstance(jsonrpc, str):
                    raise DeserializationError("field 'jsonrpc' must be of type string")

                if jsonrpc != cls.version:
                    raise DeserializationError(f"jsonrpc version '{data['jsonrpc']}' is not supported")

                request_id, error = data.get('id'), data.get('error', UNSET)
                if request_id is None and error is not UNSET:
                    return cls(responses=(), error=error_cls.from_json(error))

            if not isinstance(data, (list, tuple)):
                raise DeserializationError("data must be of type list")

        except KeyError as e:
            raise DeserializationError(f"required field {e} not found") from e

        responses = tuple(Response.from_json(item) for item in data)
        if check_ids:
            cls._check_response_id_uniqueness(responses)

        return cls(responses=responses)

    def to_json(self) -> JsonT:
        """
        Serializes the batch response to json data.

        :returns: json data
        """

        if self.is_error:
            return Response(id=None, error=self.error, result=UNSET).to_json()

        return [response.to_json() for response in self.responses or ()]

    def __getitem__(self, idx: int) -> Response:
        """
        Returns a request at index `idx`.
        """

        if not self.responses:
            raise IndexError("index out of range")

        return self.responses[idx]

    def __iter__(self) -> Iterator[Response]:
        return iter(self.responses or ())

    def __len__(self) -> int:
        return len(self.responses or ())

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, BatchResponse):
            return NotImplemented

        return all(
            it.starmap(
                op.eq, zip(
                    sorted(self, key=op.attrgetter('id')),
                    sorted(other, key=op.attrgetter('id')),
                ),
            ),
        )

    @property
    def is_success(self) -> bool:
        """
        Returns ``True`` if the response has succeeded.
        """

        return self.error is UNSET

    @property
    def is_error(self) -> bool:
        """
        Returns ``True`` if the request has not succeeded. Note that it is not the same as
        :py:attr:`pjrpc.common.BatchResponse.has_error`. `is_error` indicates that the whole batch request failed,
        whereas `has_error` indicates that one of the requests in the batch failed.
        """

        return not self.is_success

    @property
    def has_error(self) -> bool:
        """
        Returns ``True`` if any response has an error.
        """

        return any((response.is_error for response in self.responses or ()))

    def unwrap_results(self) -> tuple[MaybeSet[JsonT], ...]:
        """
        Returns the batch result as a tuple. If any response of the batch has an error
        raises an exception related to the first errored response.
        """

        if self.is_error:
            raise self.unwrap_error()

        result = []

        for response in self.responses or ():
            if response.is_error:
                raise response.unwrap_error()
            result.append(response.result)

        return tuple(result)

    def unwrap_result(self) -> JsonT:
        """
        Returns result. If result is not set raises and exception.
        """

        return self.unwrap_results()

    def unwrap_error(self) -> JsonRpcError:
        """
        Returns error. If error is not set raises and exception.
        """

        assert self.error is not UNSET, "error is not set"
        return self.error

    @classmethod
    def _check_response_id_uniqueness(cls, responses: tuple[Response, ...]) -> None:
        ids = tuple(response.id for response in responses if response.id is not None)
        if len(ids) != len(set(ids)):
            raise IdentityError("batch response ids are not unique")



import abc
import asyncio
import functools as ft
import json
import logging
import warnings
from typing import Any, Awaitable, Callable, Generic, ItemsView, Iterable, Iterator, KeysView, Optional, TypeVar, Union
from typing import ValuesView

import pjrpc
from pjrpc.common import UNSET, AbstractResponse, BatchRequest, BatchResponse, MaybeSet, Request, Response, UnsetType
from pjrpc.common.typedefs import JsonRpcParamsT
from pjrpc.server import utils
from pjrpc.server.typedefs import AsyncMiddlewareType, MiddlewareType

from . import validators

logger = logging.getLogger(__package__)


FunctionType = Callable[..., Any]
BoundMethod = Callable[[], Any]
FunctionT = TypeVar('FunctionT', bound=FunctionType)


class Method:
    """
    JSON-RPC method wrapper. Stores method itself and some metainformation.

    :param func: method function
    :param name: method name
    :param pass_context: context name
    :param validator_factory: method validator factory
    :param metadata: method metadata
    """

    def __init__(
        self,
        func: Callable[..., Any],
        name: Optional[str] = None,
        *,
        pass_context: Union[bool, str] = False,
        validator_factory: Optional[validators.BaseValidatorFactory] = None,
        metadata: Iterable[Any] = (),
    ):
        self.func = func
        self.name = name or func.__name__
        self.pass_context = pass_context
        self.metadata = list(metadata)

        if isinstance(self.pass_context, bool):
            if self.pass_context:
                exclude = utils.exclude_positional_param(0)
            else:
                exclude = None
        elif isinstance(self.pass_context, str):
            exclude = utils.exclude_named_param(self.pass_context)
        else:
            raise AssertionError("unreachable")

        validator_factory = validator_factory or validators.BaseValidatorFactory(exclude=exclude)
        self.validator = validator_factory.build(func)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.func(*args, **kwargs)

    def bind(self, params: Optional[JsonRpcParamsT], context: Optional[Any] = None) -> BoundMethod:
        method_args: list[Any] = []
        method_kwargs = self.validator.validate_params(params)

        method_args, method_kwargs = self._prepare_method_args(method_args, method_kwargs, context)

        return ft.partial(self.func, *method_args, **method_kwargs)

    def _prepare_method_args(
        self,
        args: list[Any],
        kwargs: dict[str, Any],
        context: Optional[Any],
    ) -> tuple[list[Any], dict[str, Any]]:
        if isinstance(self.pass_context, bool):
            if self.pass_context:
                args.append(context)
        elif isinstance(self.pass_context, str):
            kwargs[self.pass_context] = context
        else:
            raise AssertionError("unreachable")

        return args, kwargs


MetadataProcessorT = Callable[[Method], Method]


class MethodRegistry:
    """
    Method registry.
    """

    def __init__(
        self,
        validator_factory: Optional[validators.BaseValidatorFactory] = None,
        metadata: Iterable[Any] = (),
        metadata_processors: Iterable[MetadataProcessorT] = (),
    ) -> None:
        self._validator_factory = validator_factory
        self._metadata = list(metadata)
        self._metadata_processors = list(metadata_processors)
        self._registry: dict[str, Method] = {}

    def __iter__(self) -> Iterator[str]:
        """
        Returns registry method iterator.
        """

        return iter(self._registry)

    def __getitem__(self, item: str) -> Method:
        """
        Returns a method from the registry by name.

        :param item: method name
        :returns: found method
        :raises: KeyError
        """

        return self._registry[item]

    def items(self) -> ItemsView[str, Method]:
        return self._registry.items()

    def keys(self) -> KeysView[str]:
        return self._registry.keys()

    def values(self) -> ValuesView[Method]:
        return self._registry.values()

    def get(self, item: str) -> Optional[Method]:
        """
        Returns a method from the registry by name.

        :param item: method name
        :returns: found method or `None`
        """

        return self._registry.get(item)

    def merge(self, other: 'MethodRegistry') -> 'MethodRegistry':
        for method in other.values():
            self._add_method(method)

        return self

    def add(
        self,
        name: Optional[str] = None,
        *,
        pass_context: Union[bool, str] = False,
        metadata: Iterable[Any] = (),
    ) -> Callable[[FunctionT], FunctionT]:
        """
        Decorator adding decorated method to the registry.

        :param name: method name to be used instead of `__name__` attribute
        :param pass_context: pass application context if supported
        :param metadata: method metadata

        :returns: decorated method or decorator
        """

        def decorator(func: FunctionT) -> FunctionT:
            self._add_function(
                func,
                name,
                pass_context=pass_context,
                metadata=metadata,
            )

            return func

        return decorator

    def add_method(
        self,
        method: FunctionType,
        name: Optional[str] = None,
        *,
        pass_context: Union[bool, str] = False,
        metadata: Iterable[Any] = (),
    ) -> 'MethodRegistry':
        """
        Adds the method to the registry.

        :param method: method to be added
        :param name: method name to be used instead of `__name__` attribute
        :param pass_context: pass application context if supported
        :param metadata: method metadata

        :returns: decorated method or decorator
        """

        self._add_function(
            method,
            name,
            pass_context=pass_context,
            metadata=metadata,
        )

        return self

    def _add_function(
        self,
        func: FunctionType,
        name: Optional[str] = None,
        pass_context: Union[bool, str] = False,
        metadata: Iterable[Any] = (),
    ) -> None:
        self._add_method(
            Method(
                func,
                name,
                pass_context=pass_context,
                metadata=metadata,
                validator_factory=self._validator_factory,
            ),
        )

    def _add_method(self, method: Method) -> None:
        if method.name in self._registry:
            warnings.warn(f"method '{method.name}' already registered")

        method.metadata.extend(self._metadata)
        for metadata_processor in self._metadata_processors:
            method = metadata_processor(method)

        self._registry[method.name] = method


class JSONEncoder(pjrpc.JSONEncoder):
    """
    Server JSON encoder. All custom server encoders should be inherited from it.
    """

    def default(self, o: Any) -> Any:
        if isinstance(o, validators.base.ValidationError):
            return [err for err in o.args]

        return super().default(o)


def extract_error_codes(response: AbstractResponse) -> tuple[int, ...]:
    if isinstance(response, BatchResponse):
        return (response.error.code,) if response.error else tuple(r.error.code if r.error else 0 for r in response)
    elif isinstance(response, Response):
        return (response.error.code if response.error else 0,)
    else:
        raise AssertionError("unreachable")


class BaseDispatcher:
    """
    Method dispatcher.

    :param request_class: JSON-RPC request class
    :param response_class: JSON-RPC response class
    :param batch_request: JSON-RPC batch request class
    :param batch_response: JSON-RPC batch response class
    :param json_loader: request json loader
    :param json_dumper: response json dumper
    :param json_encoder: response json encoder
    :param json_decoder: request json decoder
    :param middlewares: request middlewares
    """

    def __init__(
        self,
        *,
        request_class: type[Request] = Request,
        response_class: type[Response] = Response,
        batch_request: type[BatchRequest] = BatchRequest,
        batch_response: type[BatchResponse] = BatchResponse,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[type[json.JSONDecoder]] = None,
        middlewares: Iterable[Callable[..., Any]] = (),
    ):
        self._json_loader = json_loader
        self._json_dumper = json_dumper
        self._json_encoder = json_encoder
        self._json_decoder = json_decoder
        self._request_class = request_class
        self._response_class = response_class
        self._batch_request = batch_request
        self._batch_response = batch_response
        self._middlewares = list(middlewares)

        self._registry = MethodRegistry()

    @property
    def registry(self) -> MethodRegistry:
        return self._registry

    def add_methods(self, registry: MethodRegistry) -> 'BaseDispatcher':
        self._registry.merge(registry)
        return self


class Executor(abc.ABC):
    """
    Method executor.
    """

    @abc.abstractmethod
    def execute(
        self,
        handler: Callable[..., Any],
        /,
        requests: Iterable[Request],
        *args: Any,
        **kwargs: Any,
    ) -> Iterable[Any]:
        pass


class BasicExecutor(Executor):
    """
    Sequential in-thread method executor
    """

    def execute(
        self,
        handler: Callable[..., Any],
        /,
        requests: Iterable[Request],
        *args: Any,
        **kwargs: Any,
    ) -> Iterable[Any]:
        return list(handler(request, *args, **kwargs) for request in requests)


ContextType = TypeVar('ContextType')


class Dispatcher(BaseDispatcher, Generic[ContextType]):
    """
    Synchronous method dispatcher.
    """

    def __init__(
        self,
        *,
        executor: Optional[Executor] = None,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[type[json.JSONDecoder]] = None,
        middlewares: Iterable['MiddlewareType[ContextType]'] = (),
        max_batch_size: Optional[int] = None,
    ):
        super().__init__(
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
        )
        self._max_batch_size = max_batch_size
        self._executor = executor or BasicExecutor()

        self._request_handler = self._wrap_handle_request()

    def dispatch(self, request_text: str, context: ContextType) -> Optional[tuple[str, tuple[int, ...]]]:
        """
        Deserializes request, dispatches it to the required method and serializes the result.

        :param request_text: request text representation
        :param context: application context (if supported)
        :return: response text representation and error codes
        """

        logger.getChild('raw_request').debug("request received: %s", request_text)

        response: MaybeSet[AbstractResponse]
        try:
            request_json = self._json_loader(request_text, cls=self._json_decoder)
            request: Union[Request, BatchRequest]
            if isinstance(request_json, (list, tuple)):
                request = self._batch_request.from_json(request_json)
            else:
                request = self._request_class.from_json(request_json)

        except json.JSONDecodeError as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.ParseError(data=str(e)))

        except (pjrpc.exceptions.DeserializationError, pjrpc.exceptions.IdentityError) as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.InvalidRequestError(data=str(e)))

        else:
            if isinstance(request, BatchRequest):
                if self._max_batch_size and len(request) > self._max_batch_size:
                    response = self._response_class(
                        id=None,
                        error=pjrpc.exceptions.InvalidRequestError(data="batch too large"),
                    )
                else:
                    responses = (
                        resp for resp in self._executor.execute(self._request_handler, request, context)
                        if not isinstance(resp, UnsetType)
                    )
                    response = self._batch_response(tuple(responses))
            else:
                response = self._request_handler(request, context)

        if not isinstance(response, UnsetType):
            response_text = self._json_dumper(response.to_json(), cls=self._json_encoder)
            logger.getChild('raw_response').debug("response prepared: %s", response_text)

            return response_text, extract_error_codes(response)

        return None

    def add_middlewares(self, *middlewares: MiddlewareType[ContextType], before: bool = False) -> None:
        if before:
            self._middlewares = list(middlewares) + self._middlewares
        else:
            self._middlewares = self._middlewares + list(middlewares)

        self._request_handler = self._wrap_handle_request()

    def _wrap_handle_request(self) -> Callable[[Request, ContextType], MaybeSet[Response]]:
        request_handler = self._handle_request
        for middleware in reversed(self._middlewares):
            request_handler = ft.partial(middleware, handler=request_handler)

        return request_handler

    def _handle_request(self, request: Request, /, context: ContextType) -> MaybeSet[Response]:
        try:
            return self._handle_rpc_request(request, context)
        except pjrpc.exceptions.JsonRpcError as e:
            logger.info("method execution error %s(%r): %r", request.method, request.params, e)
            error = e

        except Exception as e:
            logger.exception("internal server error: %r", e)
            error = pjrpc.exceptions.InternalError()

        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, error=error)

    def _handle_rpc_request(self, request: Request, context: ContextType) -> MaybeSet[Response]:
        result = self._handle_rpc_method(request.method, request.params, context)
        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, result=result)

    def _handle_rpc_method(
        self,
        method_name: str,
        params: Optional[JsonRpcParamsT],
        context: ContextType,
    ) -> Any:
        method = self._registry.get(method_name)
        if method is None:
            raise pjrpc.exceptions.MethodNotFoundError(data=f"method '{method_name}' not found")

        try:
            bound_method = method.bind(params, context=context)
        except validators.ValidationError as e:
            raise pjrpc.exceptions.InvalidParamsError(data=e) from e

        try:
            return bound_method()

        except pjrpc.exceptions.JsonRpcError:
            raise

        except Exception as e:
            logger.exception("method unhandled exception %s(%r): %r", method_name, params, e)
            raise pjrpc.exceptions.ServerError() from e


class AsyncExecutor(abc.ABC):
    """
    Asynchronous method executor.
    """

    @abc.abstractmethod
    async def execute(
        self,
        handler: Callable[..., Awaitable[Any]],
        /,
        requests: Iterable[Request],
        *args: Any,
        **kwargs: Any,
    ) -> Iterable[Any]:
        pass


class BasicAsyncExecutor(AsyncExecutor):
    """
    Sequential asynchronous method executor
    """

    async def execute(
        self,
        handler: Callable[..., Awaitable[Any]],
        /,
        requests: Iterable[Request],
        *args: Any,
        **kwargs: Any,
    ) -> Iterable[Any]:
        return [await handler(request, *args, **kwargs) for request in requests]


class ParallelAsyncExecutor(AsyncExecutor):
    """
    Parallel asynchronous method executor
    """

    async def execute(
        self,
        handler: Callable[..., Awaitable[Any]],
        /,
        requests: Iterable[Request],
        *args: Any,
        **kwargs: Any,
    ) -> Iterable[Any]:
        return list(await asyncio.gather(*(handler(request, *args, **kwargs) for request in requests)))


class AsyncDispatcher(BaseDispatcher, Generic[ContextType]):
    """
    Asynchronous method dispatcher.
    """

    def __init__(
        self,
        *,
        executor: Optional[AsyncExecutor] = None,
        json_loader: Callable[..., Any] = json.loads,
        json_dumper: Callable[..., str] = json.dumps,
        json_encoder: type[JSONEncoder] = JSONEncoder,
        json_decoder: Optional[type[json.JSONDecoder]] = None,
        middlewares: Iterable[AsyncMiddlewareType[ContextType]] = (),
        max_batch_size: Optional[int] = None,
    ):
        super().__init__(
            json_loader=json_loader,
            json_dumper=json_dumper,
            json_encoder=json_encoder,
            json_decoder=json_decoder,
            middlewares=middlewares,
        )
        self._max_batch_size = max_batch_size
        self._executor = executor or BasicAsyncExecutor()

        self._request_handler = self._wrap_handle_request()

    async def dispatch(self, request_text: str, context: ContextType) -> Optional[tuple[str, tuple[int, ...]]]:
        """
        Deserializes request, dispatches it to the required method and serializes the result.

        :param request_text: request text representation
        :param context: application context (if supported)
        :return: response text representation and error codes
        """

        logger.getChild('request').debug("request received: %s", request_text)

        response: MaybeSet[AbstractResponse]
        try:
            request_json = self._json_loader(request_text, cls=self._json_decoder)
            request: Union[Request, BatchRequest]
            if isinstance(request_json, (list, tuple)):
                request = self._batch_request.from_json(request_json)
            else:
                request = self._request_class.from_json(request_json)

        except json.JSONDecodeError as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.ParseError(data=str(e)))

        except (pjrpc.exceptions.DeserializationError, pjrpc.exceptions.IdentityError) as e:
            response = self._response_class(id=None, error=pjrpc.exceptions.InvalidRequestError(data=str(e)))

        else:
            if isinstance(request, BatchRequest):
                if self._max_batch_size and len(request) > self._max_batch_size:
                    response = self._response_class(
                        id=None,
                        error=pjrpc.exceptions.InvalidRequestError(data="batch too large"),
                    )
                else:
                    responses = (
                        resp for resp in await self._executor.execute(self._request_handler, request, context)
                        if not isinstance(resp, UnsetType)
                    )
                    response = self._batch_response(tuple(responses))
            else:
                response = await self._request_handler(request, context)

        if not isinstance(response, UnsetType):
            response_text = self._json_dumper(response.to_json(), cls=self._json_encoder)
            logger.getChild('response').debug("response sent: %s", response_text)

            return response_text, extract_error_codes(response)

        return None

    def add_middlewares(self, *middlewares: AsyncMiddlewareType[ContextType], before: bool = False) -> None:
        if before:
            self._middlewares = list(middlewares) + self._middlewares
        else:
            self._middlewares = self._middlewares + list(middlewares)

        self._request_handler = self._wrap_handle_request()

    def _wrap_handle_request(self) -> Callable[[Request, ContextType], Awaitable[MaybeSet[Response]]]:
        request_handler = self._handle_request
        for middleware in reversed(self._middlewares):
            request_handler = ft.partial(middleware, handler=request_handler)

        return request_handler

    async def _handle_request(self, request: Request, context: ContextType) -> MaybeSet[Response]:
        try:
            return await self._handle_rpc_request(request, context)
        except pjrpc.exceptions.JsonRpcError as e:
            logger.info("method execution error %s(%r): %r", request.method, request.params, e)
            error = e

        except Exception as e:
            logger.exception("internal server error: %r", e)
            error = pjrpc.exceptions.InternalError()

        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, error=error)

    async def _handle_rpc_request(self, request: Request, context: ContextType) -> MaybeSet[Response]:
        result = await self._handle_rpc_method(request.method, request.params, context)
        if request.id is None:
            return UNSET

        return self._response_class(id=request.id, result=result)

    async def _handle_rpc_method(
        self, method_name: str, params: Optional[JsonRpcParamsT], context: ContextType,
    ) -> Any:
        method = self._registry.get(method_name)
        if method is None:
            raise pjrpc.exceptions.MethodNotFoundError(data=f"method '{method_name}' not found")

        try:
            bound_method = method.bind(params, context=context)
        except validators.ValidationError as e:
            raise pjrpc.exceptions.InvalidParamsError(data=e) from e

        try:
            result = bound_method()
            if asyncio.iscoroutine(result):
                result = await result

            return result

        except pjrpc.exceptions.JsonRpcError:
            raise

        except Exception as e:
            logger.exception("method unhandled exception %s(%r): %r", method_name, params, e)
            raise pjrpc.exceptions.ServerError() from e



from typing import Any, Mapping, Optional

from pjrpc.client.client import AsyncMiddlewareHandler, MiddlewareHandler
from pjrpc.common import AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, Request, Response, exceptions


def validate_response_id_middleware(
    request: AbstractRequest,
    request_kwargs: Mapping[str, Any],
    /,
    handler: MiddlewareHandler,
) -> Optional[AbstractResponse]:
    response = handler(request, request_kwargs)
    if response is not None:
        _validate_any_response_id(request, response)

    return response


async def async_validate_response_id_middleware(
    request: AbstractRequest,
    request_kwargs: Mapping[str, Any],
    /,
    handler: AsyncMiddlewareHandler,
) -> Optional[AbstractResponse]:
    response = await handler(request, request_kwargs)
    if response is not None:
        _validate_any_response_id(request, response)

    return response


def _validate_any_response_id(request: AbstractRequest, response: AbstractResponse) -> None:
    if isinstance(request, Request) and isinstance(response, Response):
        _validate_response_id(request, response)
    elif isinstance(request, BatchRequest) and isinstance(response, BatchResponse):
        _validate_batch_response_ids(request, response)


def _validate_response_id(request: Request, response: Response) -> None:
    if response.id is not None and response.id != request.id:
        raise exceptions.IdentityError(
            f"response id doesn't match the request one: expected {request.id}, got {response.id}",
        )


def _validate_batch_response_ids(batch_request: BatchRequest, batch_response: BatchResponse) -> None:
    if batch_response.is_success:
        response_map = {response.id: response for response in batch_response if response.id is not None}

        for request in batch_request:
            if request.id is not None:
                response = response_map.pop(request.id, None)
                if response is None:
                    raise exceptions.IdentityError(f"response '{request.id}' not found")

        if response_map:
            raise exceptions.IdentityError(f"unexpected response found: {response_map.keys()}")



import inspect
import itertools as it
from typing import Any, Callable, Iterable, Optional

from pjrpc.common import UNSET, MaybeSet, UnsetType
from pjrpc.common.exceptions import TypedError

__all__ = [
    'BaseMethodInfoExtractor',
]

MethodType = Callable[..., Any]
ExcludeFunc = Callable[[int, str, Optional[type[Any]], Optional[Any]], bool]


class BaseMethodInfoExtractor:
    """
    Base method schema extractor.
    """

    def extract_params_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        """
        Extracts params schema.
        """

        return {}, {}

    def extract_request_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        """
        Extracts request schema.
        """

        return {}, {}

    def extract_result_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        """
        Extracts result schema.
        """

        return {}, {}

    def extract_response_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            errors: Optional[Iterable[type[TypedError]]] = None,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        """
        Extracts response schema.
        """

        return {}, {}

    def extract_error_response_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            errors: Optional[Iterable[type[TypedError]]] = None,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        """
        Extracts error response schema.
        """

        return {}, {}

    def extract_description(self, method: MethodType) -> MaybeSet[str]:
        """
        Extracts method description.
        """

        description: MaybeSet[str]
        if method.__doc__:
            doc = inspect.cleandoc(method.__doc__)
            description = '\n'.join(it.takewhile(lambda line: line, doc.split('\n')))
        else:
            description = UNSET

        return description

    def extract_summary(self, method: MethodType) -> MaybeSet[str]:
        """
        Extracts method summary.
        """

        description = self.extract_description(method)

        summary: MaybeSet[str]
        if not isinstance(description, UnsetType):
            summary = description.split('.')[0]
        else:
            summary = UNSET

        return summary

    def extract_errors(self, method: MethodType) -> MaybeSet[list[type[TypedError]]]:
        """
        Extracts method errors.
        """

        return UNSET

    def extract_deprecation_status(self, method: MethodType) -> MaybeSet[bool]:
        """
        Extracts method deprecation status.
        """

        return UNSET



"""
Client and server common functions, types and classes that implements JSON-RPC protocol itself
and agnostic to any transport protocol layer (http, socket, amqp) and server-side implementation.
"""

from . import exceptions, generators, typedefs
from .common import UNSET, JsonT, MaybeSet, UnsetType
from .encoder import JSONEncoder
from .exceptions import JsonRpcError
from .request import AbstractRequest, BatchRequest, Request
from .response import AbstractResponse, BatchResponse, Response

DEFAULT_CONTENT_TYPE = 'application/json'
'''default JSON-RPC client/server content type'''  # for sphinx autodoc

REQUEST_CONTENT_TYPES = ('application/json', 'application/json-rpc', 'application/jsonrequest')
'''allowed JSON-RPC server requests content types'''  # for sphinx autodoc

RESPONSE_CONTENT_TYPES = ('application/json', 'application/json-rpc')
'''allowed JSON-RPC client responses content types'''  # for sphinx autodoc


__all__ = [
    'AbstractRequest',
    'AbstractResponse',
    'BatchRequest',
    'BatchResponse',
    'DEFAULT_CONTENT_TYPE',
    'exceptions',
    'generators',
    'JSONEncoder',
    'JsonRpcError',
    'JsonT',
    'MaybeSet',
    'Request',
    'REQUEST_CONTENT_TYPES',
    'Response',
    'RESPONSE_CONTENT_TYPES',
    'typedefs',
    'UNSET',
    'UnsetType',
]



"""
Extensible `JSON-RPC <https://www.jsonrpc.org>`_ client/server library.
"""

from pjrpc.common import AbstractRequest, AbstractResponse, BatchRequest, BatchResponse, JSONEncoder, Request, Response
from pjrpc.common import exceptions, typedefs

exc = exceptions

__all__ = [
    'AbstractRequest',
    'AbstractResponse',
    'exceptions',
    'exc',
    'typedefs',
    'Request',
    'Response',
    'BatchRequest',
    'Response',
    'Response',
    'BatchResponse',
    'JSONEncoder',
]



import inspect
from typing import Any, Callable, Generic, Iterable, Literal, Optional, TypeVar, Union

import pydantic as pd
import pydantic.generics

from pjrpc.common import exceptions
from pjrpc.server.specs.extractors import BaseMethodInfoExtractor, ExcludeFunc

MethodType = Callable[..., Any]


def to_camel(string: str) -> str:
    return ''.join(word.capitalize() for word in string.split('_'))


MethodT = TypeVar('MethodT', bound=str)
ParamsT = TypeVar('ParamsT', bound=pd.BaseModel, covariant=True)


class JsonRpcRequest(pd.generics.GenericModel, Generic[MethodT, ParamsT]):
    class Config:
        title = "Request",
        extra = 'forbid'
        strict = True,

    jsonrpc: Literal['2.0', '1.0'] = pd.Field(title="Version", description="JSON-RPC protocol version")
    id: Optional[Union[str, int]] = pd.Field(None, description="Request identifier", examples=[1])
    method: MethodT = pd.Field(description="Method name")
    params: ParamsT = pd.Field(description="Method parameters")


ResultT = TypeVar('ResultT', covariant=True)


class JsonRpcResponseSuccess(pd.generics.GenericModel, Generic[ResultT]):
    class Config:
        title = 'Success',
        extra = 'forbid'
        strict = True,

    jsonrpc: Literal['2.0', '1.0'] = pd.Field(title="Version", description="JSON-RPC protocol version")
    id: Union[str, int] = pd.Field(description="Request identifier", examples=[1])
    result: ResultT = pd.Field(description="Method execution result")


ErrorCodeT = TypeVar('ErrorCodeT', bound=int)
ErrorDataT = TypeVar('ErrorDataT')


class JsonRpcError(pd.generics.GenericModel, Generic[ErrorCodeT, ErrorDataT]):
    class Config:
        title = "Error",
        extra = 'forbid'
        strict = True,

    code: ErrorCodeT = pd.Field(description="Error code")
    message: str = pd.Field(description="Error message")
    data: ErrorDataT = pd.Field(description="Error additional data")


JsonRpcErrorT = TypeVar('JsonRpcErrorT', bound=JsonRpcError[Any, Any], covariant=True)


class JsonRpcResponseError(pd.generics.GenericModel, Generic[JsonRpcErrorT]):
    class Config:
        title = "ResponseError",
        extra = 'forbid'
        strict = True

    jsonrpc: Literal['1.0', '2.0'] = pd.Field(title="Version", description="JSON-RPC protocol version")
    id: Union[str, int] = pd.Field(description="Request identifier", examples=[1])
    error: JsonRpcErrorT = pd.Field(description="Request error")


class PydanticMethodInfoExtractor(BaseMethodInfoExtractor):
    """
    Pydantic method specification extractor.

    :param config_args: model configuration parameters
    :param exclude: a function that decides if the parameters must be excluded
                    from schema (useful for dependency injection)
    """

    def __init__(self, exclude: Optional[ExcludeFunc] = None, **config_args: Any):
        self._exclude = exclude
        self._config_args = config_args

    def extract_params_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        params_model = self._build_params_model(method_name, method)
        params_schema = params_model.schema(ref_template=ref_template)

        return params_schema, params_schema.pop('definitions', {})

    def extract_request_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        params_model = self._build_params_model(method_name, method)

        # create inner model to set a reference to the reqeust model, not the model itself
        class RequestModel(pydantic.BaseModel):
            Config = pydantic.config.get_config(dict(self._config_args, title=f"{to_camel(method_name)}Request"))
            __root__: JsonRpcRequest[Literal[method_name], params_model]  # type: ignore[valid-type]

        request_model = pd.create_model(f'{to_camel(method.__name__)}Request', __base__=RequestModel)
        request_schema = request_model.schema(ref_template=ref_template)

        return request_schema, request_schema.pop('definitions', {})

    def extract_result_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        result_model = self._build_result_model(method_name, method)
        result_schema = result_model.schema(ref_template=ref_template)

        return result_schema, result_schema.pop('definitions', {})

    def extract_response_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            errors: Optional[Iterable[type[exceptions.TypedError]]] = None,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        return_model = self._build_result_model(method_name, method)

        response_model: type[pd.BaseModel]
        error_models = tuple(
            pd.create_model(
                error.__name__,
                __base__=JsonRpcResponseError[JsonRpcError[Literal[error.CODE], Any]],  # type: ignore[name-defined]
                __config__=pydantic.config.get_config(
                    dict(
                        self._config_args,
                        title=error.__name__,
                        json_schema_extra=dict(description=f'**{error.CODE}** {error.MESSAGE}'),
                    ),
                ),
            ) for error in errors or []
        )

        class ResponseModel(pydantic.BaseModel):
            Config = pydantic.config.get_config(dict(title=f"{to_camel(method_name)}Response"))
            __root__: Union[(JsonRpcResponseSuccess[return_model], *error_models)]  # type: ignore[valid-type]

        response_model = pd.create_model(f'{to_camel(method.__name__)}Response', __base__=ResponseModel)
        response_schema = response_model.schema(ref_template=ref_template)
        if error_models:
            response_schema['description'] = '\n'.join(
                f'* {error.schema().get("description", error.__name__)}'
                for error in error_models
            )

        return response_schema, response_schema.pop('definitions', {})

    def extract_error_response_schema(
            self,
            method_name: str,
            method: MethodType,
            ref_template: str,
            errors: Optional[Iterable[type[exceptions.TypedError]]] = None,
    ) -> tuple[dict[str, Any], dict[str, dict[str, Any]]]:
        error_models: list[type[pd.BaseModel]] = []
        for error in errors or []:
            class ErrorModel(pydantic.BaseModel):
                Config = pydantic.config.get_config(
                    dict(
                        self._config_args,
                        title=error.__name__,
                        json_schema_extra=dict(description=f'**{error.CODE}** {error.MESSAGE}'),
                    ),
                )
                __root__: JsonRpcResponseError[JsonRpcError[Literal[error.CODE], Any]]  # type: ignore[name-defined]

            error_models.append(pd.create_model(error.__name__, __base__=ErrorModel))

        class ResponseModel(pydantic.BaseModel):
            Config = pydantic.config.get_config(dict(title=f"{to_camel(method_name)}Response"))
            __root__: Union[tuple(error_models)]  # type: ignore[valid-type]

        response_model = pd.create_model(f'{to_camel(method.__name__)}Response', __base__=ResponseModel)
        response_schema = response_model.schema(ref_template=ref_template)

        if error_models:
            response_schema['description'] = '\n'.join(
                f'* {error.schema().get("description", error.__name__)}'
                for error in error_models
            )
        else:
            response_schema['description'] = 'Error'

        return response_schema, response_schema.pop('definitions', {})

    def _build_params_model(
            self,
            method_name: str,
            method: MethodType,
    ) -> type[pd.BaseModel]:
        signature = inspect.signature(method)

        field_definitions: dict[str, Any] = {}
        for idx, param in enumerate(signature.parameters.values()):
            if self._exclude and self._exclude(idx, param.name, param.annotation, param.default):
                continue

            if param.kind in [inspect.Parameter.POSITIONAL_OR_KEYWORD, inspect.Parameter.KEYWORD_ONLY]:
                field_definitions[param.name] = (
                    param.annotation if param.annotation is not inspect.Parameter.empty else Any,
                    param.default if param.default is not inspect.Parameter.empty else ...,
                )

        return pd.create_model(
            f'{to_camel(method.__name__)}Parameters',
            **field_definitions,
            __config__=pydantic.config.get_config(
                dict(self._config_args, title=f"{to_camel(method_name)}Parameters", extra='forbid'),
            ),
        )

    def _build_result_model(self, method_name: str, method: MethodType) -> type[pd.BaseModel]:
        result = inspect.signature(method)

        if result.return_annotation is inspect.Parameter.empty:
            return_annotation = Any
        elif result.return_annotation is None:
            return_annotation = None
        else:
            return_annotation = result.return_annotation

        class ResultModel(pd.BaseModel):
            Config = pydantic.config.get_config(dict(self._config_args, title=f"{to_camel(method_name)}Result"))
            __root__: return_annotation  # type: ignore[valid-type]

        return pd.create_model(f'{to_camel(method.__name__)}Result', __base__=ResultModel)
